#!/usr/bin/env python3

# estimate_spacing.py calls routines contained in spacing.py to calculate wing
# scale measurements from 1D SAXS data.
# It takes a list of sample file paths as input and writes estimates of scale
# features, measurements of variability and error to a tab separated file
# The code is not guaranteed to be bug free

# Add permisions to run or call with Python 3 interpreter
# estimate_spacing.py <paths_to_sample_files.txt> <species> <outfile.txt>

# Author: Juan Enciso-Romero - jenciso1@sheffield.ac.uk
# University of Sheffield
# Universidad del Rosario

import spacing
import pandas as pd
import numpy as np
from sys import argv, exit

species = argv[2].lower()
if species != "erato" and species != "melpomene":
    exit("Stopping execution: Species must be either erato or melpomene")

# Cycle through all files in the list
with open(argv[3], 'w') as current_outfile, open(argv[1], 'r') as current_paths:
    #
    current_outfile.write("ind_name\tridge_spacing\tcrossrib_spacing\tridge_std\tcrossrib_std\tcoef_var_ridge\tcoef_var_crossrib\tridge_sem\tcrossrib_sem\n")
    for line in current_paths.readlines():
        wing_data = pd.read_csv(line.strip(), sep="\t", header=None, names=["Source_file", "Ind_name", "Frame", "Scattering", "Intensity", "Angle"])
        # Normalise positions based on order encountered in Frame col.
        wing_data['Position'] = 1 + wing_data.Frame - np.min(wing_data.Frame)
        ind_name = str(np.unique(wing_data['Ind_name'].values)[0]).split("_")[0]
        print("Analysing {}".format(ind_name))
        # Create ln cols for scattering and intensity, just in case
        wing_data['Log_Scattering'] = wing_data.Scattering.apply(np.log)
        wing_data['Log_Intensity'] = wing_data.Intensity.apply(np.log)

        # Drop rows with nan and scattering values below 0.004
        wing_data = wing_data.dropna()

        wing_data = wing_data.loc[wing_data['Scattering'] > 0.00524]
        wing_data = wing_data.loc[wing_data['Scattering'] < 0.018]

        # Get only data that has angle 0
        wing_data = wing_data.loc[wing_data['Angle'] == 0]

        # Max and min points
        maxes, mins = spacing.find_max_and_min(wing_data, species)

        # Ridge spacing and crossrib spacing
        ridge_space, crossrib_space, ridge_space_std, crossrib_space_std, ridge_space_sem, crossrib_space_sem = spacing.find_spacing_estimates(wing_data, maxes, mins)
        current_outfile.write("{}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\t{:.4f}\n".format(ind_name, ridge_space, crossrib_space, ridge_space_std, crossrib_space_std, ridge_space_std/ridge_space, crossrib_space_std/crossrib_space, ridge_space_sem, crossrib_space_sem))
