# spacing.py contains routines called by estimate_spacing.py to calculate wing
# scale measurements from 1D SAXS data. The code is not guaranteed to be bug free

# Author: Juan Enciso-Romero - jenciso1@sheffield.ac.uk
# University of Sheffield
# Universidad del Rosario

import pandas as pd
import numpy as np
from scipy import signal, stats
import lmfit


# Helper fns
def get_line_params(p_x, p_y):  # p_x and p_y are arrays. -> (float, float).
    slope, intercept, r_value, p_value, std_err = stats.linregress(p_x, p_y)
    return(slope, intercept)


# Fns
def find_max_and_min(par_df, par_spp):
    ind_name = str(np.unique(par_df['Ind_name'].values)[0]).split("_")[0]
    unique_frames = np.unique(par_df['Frame'].values)
    outside_list_second_bit = []
    outside_list_first_bit = []
    unique_maxes = {}
    non_unique_maxes = {}

    # HARD_FB_LIMIT MELPOMENE 0.0092
    # HARD_SB_LIMIT MELPOMENE 0.0085
    # HARD_FB_LIMIT ERATO 0.0101
    # HARD_SB_LIMIT ERATO 0.0093
    if par_spp == "erato":
        HARD_FB_LIMIT = 0.0101
        HARD_SB_LIMIT = 0.0093
    elif par_spp == "melpomene":
        HARD_FB_LIMIT = 0.0092
        HARD_SB_LIMIT = 0.0085

    for i in np.nditer(unique_frames, flags=['buffered'], op_dtypes=['int64']):
        cast_idx = int(i)
        my_subset = par_df[par_df['Frame'] == cast_idx]
        this_scattering = my_subset['Log_Scattering'].values
        this_intensity = my_subset['Log_Intensity'].values

        # Indices of subsets
        first_bit_idcs = np.nonzero(this_scattering < np.log(HARD_FB_LIMIT))[0]
        second_bit_idcs = np.nonzero(this_scattering > np.log(HARD_SB_LIMIT))[0]
        # Need a scaler for second bit indices
        second_bit_scaler = np.min(second_bit_idcs)

        # Get subsets of scattering
        first_bit_scattering = this_scattering[np.nonzero(this_scattering < np.log(HARD_FB_LIMIT))[0]]
        second_bit_scattering = this_scattering[np.nonzero(this_scattering > np.log(HARD_SB_LIMIT))[0]]
        # And of intensity
        first_bit_intensity = this_intensity[np.nonzero(this_scattering < np.log(HARD_FB_LIMIT))[0]]
        second_bit_intensity = this_intensity[np.nonzero(this_scattering > np.log(HARD_SB_LIMIT))[0]]

        # Decide whether some bits need to be added to bad_dict
        needs_uniquifying = False

        # Find max values
        try:
            trsz_second = 9
            my_second_smooth = signal.savgol_filter(second_bit_intensity, trsz_second, 7)
            my_second_max = signal.argrelmax(my_second_smooth)
        except TypeError as type_err:
            print("Problem found when trying to apply SavGol filter for the first time to second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, type_err))
            continue
        except ValueError as val_err:
            print("Problem found when trying to apply SavGol filter for the first time to second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, val_err))
            continue

        while my_second_max[0].size > 1:
            trsz_second += 2
            try:
                my_second_smooth = signal.savgol_filter(second_bit_intensity, trsz_second, 7)
                my_second_max = signal.argrelmax(my_second_smooth)  # Max Indices
            except TypeError as type_err:
                # print("Could not find a unique maximum for second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, type_err))
                break
            except ValueError as val_err:
                # print("Could not find a unique maximum for second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, val_err))
                break

        # Rescale
        rescaled_second = my_second_max[0] + second_bit_scaler

        # Append scattering depending on whether only a unique max point was found.
        if my_second_max[0].size == 1:
            outside_list_second_bit.append(second_bit_scattering[my_second_max[0][0]])
        elif my_second_max[0].size > 1:
            needs_uniquifying = True

        # Find maximum values
        try:
            trsz_first = 9
            my_first_smooth = signal.savgol_filter(first_bit_intensity, trsz_first, 7)
            my_first_max = signal.argrelmax(my_first_smooth)
        except TypeError as type_err:
            print("Problem found when trying to apply SavGol filter for the first time to first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, type_err))
            continue
        except ValueError as val_err:
            print("Problem found when trying to apply SavGol filter for the first time to first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, val_err))
            continue

        while my_first_max[0].size > 1:
            trsz_first += 2
            try:
                my_first_smooth = signal.savgol_filter(first_bit_intensity, trsz_first, 7)
                my_first_max = signal.argrelmax(my_first_smooth)  # Max Indices
            except TypeError as type_err:
                # print("Could not find a unique maximum for first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, type_err))
                break
            except ValueError as val_err:
                # print("Could not find a unique maximum for first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, cast_idx, val_err))
                break

        # Append scattering depending on whether only a unique max point was found.
        if my_first_max[0].size == 1:
            outside_list_first_bit.append(first_bit_scattering[my_first_max[0][0]])
        elif my_first_max[0].size > 1:
            needs_uniquifying = True

        if needs_uniquifying:
            non_unique_maxes[cast_idx] = (my_first_max[0], rescaled_second, this_scattering[my_first_max[0]], this_scattering[rescaled_second])
        else:
            unique_maxes[cast_idx] = (my_first_max[0], rescaled_second)  # Do we need something else in this tuple?

        # LOOP ENDS

    # Compute mean and spread
    mean_first_bit = np.mean(np.array(outside_list_first_bit))
    mean_second_bit = np.mean(np.array(outside_list_second_bit))

    # Add results to corrected maxes dict
    corrected_maxes = {}
    for frame in non_unique_maxes:
        # Evaluate scattering
        try:
            unique_fid = np.abs(non_unique_maxes[frame][2] - mean_first_bit).argmin()
        except ValueError as val_err:
            print("Skipping frame {} from individual {} on first bit while trying to uniquify maxima.\nMessage: {}".format(frame, ind_name, val_err))
            continue
        try:
            unique_sid = np.abs(non_unique_maxes[frame][3] - mean_second_bit).argmin()
        except ValueError as val_err:
            print("Skipping frame {} from individual {} on second bit while trying to uniquify maxima.\nMessage: {}".format(frame, ind_name, val_err))
            continue
        corrected_maxes[frame] = (np.array([non_unique_maxes[frame][0][unique_fid],]), np.array([non_unique_maxes[frame][1][unique_sid],]))
        # XXX: RESCALE

    merged_maxes = {}
    merged_maxes.update(unique_maxes)
    merged_maxes.update(corrected_maxes)

    bad_max_frames = [i for i in merged_maxes if merged_maxes[i][0].size == 0 or merged_maxes[i][1].size == 0 or merged_maxes[i][0][0] > merged_maxes[i][1][0]]

    for i in bad_max_frames:
        del merged_maxes[i]

    # Find minima

    unique_mins = {}
    non_unique_mins = {}
    helper_list_first_mins = []
    helper_list_second_mins = []

    for j in merged_maxes:
        my_subset = par_df[par_df['Frame'] == j]
        this_scattering = my_subset['Log_Scattering'].values
        this_intensity = my_subset['Log_Intensity'].values

        # Indices of subsets
        first_bit_idcs = np.arange(0, merged_maxes[j][0] + 1)
        second_bit_idcs = np.arange(merged_maxes[j][0], merged_maxes[j][1] + 1)
        second_bit_scaler = np.min(second_bit_idcs)

        # Get subsets of scattering
        first_bit_scattering = this_scattering[first_bit_idcs]
        second_bit_scattering = this_scattering[second_bit_idcs]
        # And of intensity
        first_bit_intensity = this_intensity[first_bit_idcs]
        second_bit_intensity = this_intensity[second_bit_idcs]

        needs_uniquifying = False

        # Find maximum values
        try:
            trsz_second = 9
            my_second_smooth = signal.savgol_filter(second_bit_intensity, trsz_second, 7)
            my_second_min = signal.argrelmin(my_second_smooth)
        except TypeError as type_err:
            print("Problem found when trying to apply SavGol filter for the first time to second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, type_err))
            continue
        except ValueError as val_err:
            print("Problem found when trying to apply SavGol filter for the first time to second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, val_err))
            continue

        while my_second_min[0].size > 1:
            trsz_second += 2
            try:
                my_second_smooth = signal.savgol_filter(second_bit_intensity, trsz_second, 7)
                my_second_min = signal.argrelmin(my_second_smooth)  # Max Indices
            except TypeError as type_err:
                # print("Could not find a unique maximum for second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, type_err))
                break
            except ValueError as val_err:
                # print("Could not find a unique maximum for second bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, val_err))
                break

        #  Rescale
        rescaled_second = my_second_min[0] + second_bit_scaler

        # Append scattering depending on whether only a unique max point was found.
        if my_second_min[0].size == 1:
            helper_list_second_mins.append(second_bit_scattering[my_second_min[0][0]])
        elif my_second_min[0].size > 1:
            needs_uniquifying = True

        # Find maxes - Go with first bit now
        try:
            trsz_first = 9
            my_first_smooth = signal.savgol_filter(first_bit_intensity, trsz_first, 7)
            my_first_min = signal.argrelmin(my_first_smooth)
        except TypeError as type_err:
            print("Problem found when trying to apply SavGol filter for the first time to first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, type_err))
            continue
        except ValueError as value_err:
            print("Problem found when trying to apply SavGol filter for the first time to first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, value_err))
            continue

        while my_first_min[0].size > 1:
            trsz_first += 2
            try:
                my_first_smooth = signal.savgol_filter(first_bit_intensity, trsz_first, 7)
                my_first_min = signal.argrelmin(my_first_smooth)  # Max Indices
            except TypeError as type_err:
                # print("Could not find a unique maximum for first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, type_err))
                break
            except ValueError as val_err:
                # print("Could not find a unique maximum for first bit.\nIndividual: {}. Frame: {}.\nMessage: {}".format(ind_name, j, val_err))
                break

        # Append scattering depending on whether only a unique min point was found.
        if my_first_min[0].size == 1:
            helper_list_first_mins.append(first_bit_scattering[my_first_min[0][0]])
        elif my_first_min[0].size > 1:
            needs_uniquifying = True

        if needs_uniquifying:
            non_unique_mins[j] = (my_first_min[0], rescaled_second, this_scattering[my_first_min[0]], this_scattering[rescaled_second])
        else:
            unique_mins[j] = (my_first_min[0], rescaled_second)  # Do we need something else in this tuple?

        # END OF LOOP

    # UNIQUIFY
    # Compute mean and spread
    mean_first_bit_mins = np.mean(np.array(helper_list_first_mins))
    mean_second_bit_mins = np.mean(np.array(helper_list_second_mins))

    # Define a uniquified dict and add results to it
    corrected_mins = {}

    for frame in non_unique_mins:
        # Evaluate scattering
        try:
            unique_fid = np.abs(non_unique_mins[frame][2] - mean_first_bit_mins).argmin()
        except ValueError as val_err:
            print("Skipping frame {} from individual {} on first bit while trying to uniquify minima\nMessage: {}.".format(frame, ind_name, val_err))
            continue
        try:
            unique_sid = np.abs(non_unique_mins[frame][3] - mean_second_bit_mins).argmin()
        except ValueError as val_err:
            print("Skipping frame {} from individual {} on second bit while trying to uniquify minima\nMessage: {}.".format(frame, ind_name, val_err))
            continue
        corrected_mins[frame] = (np.array([non_unique_mins[frame][0][unique_fid],]), np.array([non_unique_mins[frame][1][unique_sid],]))
        # XXX: RESCALE

    merged_mins = {}
    merged_mins.update(unique_mins)
    merged_mins.update(corrected_mins)

    bad_min_frames = [i for i in merged_mins if merged_mins[i][0].size == 0 or merged_mins[i][1].size == 0]

    for i in bad_min_frames:
        del merged_mins[i]

    return (merged_maxes, merged_mins)


def find_spacing_estimates(par_df, par_maxes, par_mins):
    ind_name = str(np.unique(par_df['Ind_name'].values)[0]).split("_")[0]
    # unique_positions = np.unique(par_df['Position'].values)
    unique_frames = np.unique(par_df['Frame'].values)

    q_ridge_estimates = []
    q_crossrib_estimates = []

    for i in np.nditer(unique_frames, flags=['buffered'], op_dtypes=['int64']):
        cast_idx = int(i)
        my_subset = par_df[par_df['Frame'] == cast_idx]
        this_scattering = my_subset['Log_Scattering'].values
        this_intensity = my_subset['Log_Intensity'].values

        try:
            # Get subsets of scattering based on par_mins
            ridge_scattering = this_scattering[np.arange(par_mins[cast_idx][0], par_mins[cast_idx][1])]
            crossrib_scattering = this_scattering[np.arange(par_mins[cast_idx][1], this_scattering.size)]
            # And of intensity based on par_mins
            ridge_intensity = this_intensity[np.arange(par_mins[cast_idx][0], par_mins[cast_idx][1])]
            crossrib_intensity = this_intensity[np.arange(par_mins[cast_idx][1], this_intensity.size)]

            # Get line parameters
            ridge_slope, ridge_intercept = get_line_params(ridge_scattering[[0, -1]], ridge_intensity[[0, -1]])
            crossrib_slope, crossrib_intercept = get_line_params(crossrib_scattering[[0, -1]], crossrib_intensity[[0, -1]])
        except (KeyError, TypeError):
            print("Problem detected when trying to elaborate fitting parameters, skipping frame {} for individual {}.".format(cast_idx, ind_name))
            continue

        ############################################################################
        # Fit models using lmfit.models.LinearModel and lmfit.models.LorentzianModel
        ############################################################################

        # RIDGE
        ridge_LM = lmfit.models.LinearModel(prefix="ridgeLM_{}_".format(cast_idx))
        ridge_Lorentzian = lmfit.models.LorentzianModel(prefix="ridgeLorentzian_{}_".format(cast_idx))
        ridge_pars = ridge_LM.make_params(intercept=ridge_intercept, slope=ridge_slope)
        ridge_pars += ridge_Lorentzian.make_params(amplitude=2/(np.pi*this_intensity[par_maxes[cast_idx][0][0]]), center=this_scattering[par_maxes[cast_idx][0][0]], sigma=1/(np.pi*this_intensity[par_maxes[cast_idx][0][0]]))

        combined_ridge = ridge_LM+ridge_Lorentzian
        ridge_out = combined_ridge.fit(ridge_intensity, ridge_pars, x=ridge_scattering)

        # CROSSRIB
        crossrib_LM = lmfit.models.LinearModel(prefix="crossribLM_{}_".format(cast_idx))
        crossrib_Lorentzian = lmfit.models.LorentzianModel(prefix="crossribLorentzian_{}_".format(cast_idx))
        crossrib_pars = crossrib_LM.make_params(intercept=crossrib_intercept, slope=crossrib_slope)
        crossrib_pars += crossrib_Lorentzian.make_params(amplitude=2/(np.pi*this_intensity[par_maxes[cast_idx][1][0]]), center=this_scattering[par_maxes[cast_idx][1][0]], sigma=1/(np.pi*this_intensity[par_maxes[cast_idx][1][0]]))

        combined_crossrib = crossrib_LM+crossrib_Lorentzian
        crossrib_out = combined_crossrib.fit(crossrib_intensity, crossrib_pars, x=crossrib_scattering)

        if ridge_out.params["ridgeLorentzian_{}_center".format(cast_idx)].stderr < 0.01:
            q_ridge_estimates.append(ridge_out.params["ridgeLorentzian_{}_center".format(cast_idx)].value)

        if crossrib_out.params["crossribLorentzian_{}_center".format(cast_idx)].stderr < 0.01:
            q_crossrib_estimates.append(crossrib_out.params["crossribLorentzian_{}_center".format(cast_idx)].value)
        # LOOP ENDS

    ridge_spacing_estimates = list(map(lambda x: (2*np.pi)/np.exp(x), q_ridge_estimates))
    crossrib_spacing_estimates = list(map(lambda x: (2*np.pi)/np.exp(x), q_crossrib_estimates))

    ridge_spacing_avg = np.mean(np.array(ridge_spacing_estimates))
    crossrib_spacing_avg = np.mean(np.array(crossrib_spacing_estimates))
    ridge_spacing_std = np.std(np.array(ridge_spacing_estimates))
    crossrib_spacing_std = np.std(np.array(crossrib_spacing_estimates))
    ridge_spacing_sem = ridge_spacing_std/np.sqrt(np.array(ridge_spacing_estimates).size)
    crossrib_spacing_sem = crossrib_spacing_std/np.sqrt(np.array(crossrib_spacing_estimates).size)

    return(ridge_spacing_avg, crossrib_spacing_avg, ridge_spacing_std, crossrib_spacing_std, ridge_spacing_sem, crossrib_spacing_sem)
